/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef GLOBUS_GRIDFTP_SERVER_S3_BASE_H
#define GLOBUS_GRIDFTP_SERVER_S3_BASE_H 1

/**
 * @file globus_gridftp_server_s3.h
 * Interfaces used to implement the s3 DSI
 */
#include "globus_gridftp_server.h"
#include "libs3.h"

#include <stdbool.h>
#include <openssl/md5.h>

#define GLOBUS_S3_CONFIG_FILE_STRING_MAX 256

extern globus_module_descriptor_t globus_i_s3_base_module;
#define GLOBUS_GRIDFTP_SERVER_S3_BASE_MODULE (&globus_i_s3_base_module)

extern char                            *globus_i_s3_default_user_agent;
extern char                            *globus_i_s3_default_hostname;
extern char                            *globus_i_s3_mime_type_file;
extern char                            *globus_i_s3_mime_types;
extern globus_hashtable_t               globus_i_s3_mime_extensions;
extern bool                             globus_i_s3_https;
extern bool                             globus_i_s3_verify_peer;
extern bool                             globus_i_s3_use_virtual_host;
extern int                              globus_i_s3_max_retries;
extern int                              globus_i_s3_retry_delay_ms;

#define GLOBUS_S3_MTIME_NAME            "mtime"
#define GLOBUS_S3_MD5_NAME              "md5"
#define GLOBUS_S3_DIR_PLACEHOLDER_TYPE  "application/x-globus-dir-placeholder"

#define GLOBUS_S3_BUCKET_CONTEXT(handle, bucket_name) \
        (S3BucketContext) \
        { \
            .bucketName           = (bucket_name), \
            .protocol             = (handle)->protocol, \
            .uriStyle             = (handle)->uriStyle, \
            .accessKeyId          = (handle)->accessKeyId, \
            .secretAccessKey      = (handle)->secretAccessKey, \
            .securityToken        = (handle)->securityToken \
        }

typedef struct
{
    char                                extension[32];
    char                                mime_type[];
}
globus_i_s3_mime_t;

typedef struct
globus_i_s3_config_key_s
{
    enum
    {
        GLOBUS_S3_CONFIG_STRING,
        GLOBUS_S3_CONFIG_BOOL,
        GLOBUS_S3_CONFIG_INT
    }
    value_type;

    const char                         *key_name;
    char                              **value_string;
    bool                               *value_bool;
    int                               *value_int;
    bool                                required;
}
globus_i_s3_config_key_t;

globus_result_t
globus_i_s3_config_parse(
    const char                         *config_path,
    globus_i_s3_config_key_t           *extra_config_keys);

typedef
struct globus_i_s3_buffer_s
{
    /** Size of the buffer array */
    size_t                              buffer_len;
    /** Amount of initialized data in the buffer array */
    size_t                              buffer_used;
    /** Transfer offset of this buffer in a recv */
    uint64_t                            transfer_offset;
    /** Next buffer in a list */
    struct globus_i_s3_buffer_s        *next;
    /** Buffer data block */
    unsigned char                       buffer[];
}
globus_i_s3_buffer_t;

typedef struct
{
    /** Etag of the part after it has been completed */
    char                                etag[MD5_DIGEST_LENGTH*2+3];
    /** Offset into the transfer of the start of this part */
    uint64_t                            part_offset;
    /** Length of this part */
    uint64_t                            part_length;
}
globus_i_s3_part_t;


typedef
struct globus_i_s3_transfer_s
{
    /** Info from gridftp server about transfer */
    globus_gfs_transfer_info_t         *transfer_info;
    /** Final status of the operation, may be used for retry decisions */
    S3Status                            status;
    /** Number of retries for the current operation. */
    int                                 retries;
    /** Size of the transfer */
    uint64_t                            size;
    /** Current offset of an in-progress transfer */
    uint64_t                            offset;
    /**
     * True if the read callback has eof flag set on it. This means
     * we will no longer register reads and continue processing until we've
     * sent it all to s3.
     */
    bool                                eof;
    /**
     * List of buffers which have been registered for read or write
     */
    globus_i_s3_buffer_t               *registered_buffers;
    /**
     * List of full buffers which haven't yet been registered for writing to
     * s3 or the gridftp server
     */
    globus_i_s3_buffer_t               *pending_buffers;
    /**
     * List tail of pending_buffers.
     */
    globus_i_s3_buffer_t              **pending_buffers_last;
    /**
     * Current partially-filled buffer. [SEND ONLY]
     */
    globus_i_s3_buffer_t               *current_buffer;
    /**
     * List of unused buffers.
     */
    globus_i_s3_buffer_t               *free_buffers;
    /** GridFTP operation */
    globus_gfs_operation_t              op;
    /** Bucket Context derived from the session info and the pathname */
    S3BucketContext                     bucket_context;
    /** pointer into pathname that contains the bucket name */
    char                               *bucket_name;
    /** pointer into pathname that contains the key name */
    char                               *key;
    /** Result of the operation */
    globus_result_t                     result;
    globus_mutex_t                      mutex;
    globus_cond_t                       cond;

    /** True if this is a multipart restart */
    bool                                multipart_restart;
    /** Number of parts for a multipart recv [MULTIPART RECV ONLY] */
    size_t                              num_parts;
    /** If known, the size of the file being uploaded [MULTIPART RECV ONLY] */
    uint64_t                            total_file_size;
    /** The size of a part of a multipart upload [MULTIPART RECV ONLY] */
    uint64_t                            part_size;
    /** Id string of a multipart transfer [MULTIPART RECV ONLY] */
    char                               *multipart_id;
    /** Array of num_parts elements containing info about individual parts [MULTIPART RECV ONLY] */
    globus_i_s3_part_t                 *parts;
    /** Flag indicating that the multipart complete is done, so return 0 from
     * the data handler [MULTIPART RECV ONLY]
     */
    bool                                complete_end;
    /** Next marker when listing the parts of a multipart upload */
    char                               *next_marker;
    
    /** Copy of the pathname string split into bucket_name and key */
    char                                pathname[];
}
globus_i_s3_transfer_t;

typedef
struct globus_i_s3_handle_s
{
    S3Protocol                          protocol;
    S3UriStyle                          uriStyle;
    char                               *accessKeyId;
    char                               *secretAccessKey;
    char                               *securityToken;
}
globus_i_s3_handle_t;

extern
int 
globus_i_s3_base_activate(void);

extern
int 
globus_i_s3_base_deactivate(void);

extern
void
globus_i_s3_destroy(
    void                               *user_arg);

extern
void
globus_i_s3_stat(
    globus_gfs_operation_t              op,
    globus_gfs_stat_info_t             *stat_info,
    void                               *user_arg);

extern
void
globus_i_s3_command(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg);

extern
void
globus_i_s3_mkdir(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg);

extern
void
globus_i_s3_rmdir(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg);

extern
void
globus_i_s3_rename(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg);

extern
void
globus_i_s3_delete(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg);

extern
void
globus_i_s3_path_parse(
    char                               *path,
    char                              **bucket_name,
    char                              **key);

extern
void
globus_i_s3_recv(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    void                               *user_arg);

extern
bool
globus_i_s3_recv_is_reading_complete(
    const globus_i_s3_transfer_t       *transfer_state);

extern
bool
globus_i_s3_recv_is_transfer_offset_ready(
    const globus_i_s3_transfer_t       *transfer_state);

extern
globus_result_t
globus_i_s3_recv_multipart(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    uint64_t                            transfer_size,
    uint64_t                            part_size,
    bool                                restart,
    void                               *user_arg);

extern
globus_result_t
globus_i_s3_recv_multipart_abort(
    globus_i_s3_transfer_t             *transfer_state);

extern
globus_result_t
globus_i_s3_recv_multipart_list_parts(
    globus_i_s3_transfer_t             *transfer_state);

extern
globus_result_t
globus_i_s3_recv_multipart_resolve_id(
    globus_i_s3_transfer_t             *transfer_state);

extern
globus_result_t
globus_i_s3_recv_single(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    void                               *user_arg);

extern
void
globus_i_s3_send(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    void                               *user_arg);

extern
globus_i_s3_buffer_t *
globus_i_s3_buffer_get(
    globus_i_s3_transfer_t             *transfer_info,
    size_t                              size);

extern
void
globus_i_s3_buffer_release(
    globus_i_s3_transfer_t             *transfer_state,
    globus_i_s3_buffer_t               *s3_buffer);

int
globus_i_s3_mime_types_parse(
    const char                         *mime_types);

const char *
globus_i_s3_mime_type_get(
    const char                         *filename);

globus_result_t
globus_i_s3_md5sum_encode(
    const char                          md5sum[static 32],
    char                                out[static 25]);

extern
void
globus_i_s3_retry_delay(
    unsigned int                        retry_number);

void
globus_i_s3_checksum(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg);

enum
{
    GLOBUS_GFS_ERROR_S3 = 1
};

#define GlobusGFSErrorS3(status, path, details) \
    globus_error_put(GlobusGFSErrorS3Object(status, path, details))
#define GlobusGFSErrorS3Object(status, path, details) \
    globus_error_construct_error( \
        GLOBUS_GRIDFTP_SERVER_S3_BASE_MODULE, \
        NULL, \
        GLOBUS_GFS_ERROR_S3, \
        __FILE__, \
        __func__, \
        __LINE__, \
        "S3 Error accessing \"%s\": %s: %s", \
        (details && details->resource) ? details->resource : (path?path:""), \
        S3_get_status_name(status), \
        (details && details->message) ? details->message : S3_get_status_name(status))

#define _gfs_name __func__

GlobusDebugDeclare(GLOBUS_S3);

enum
{
    GLOBUS_S3_DEBUG_TRACE             = 1,
    GLOBUS_S3_DEBUG_INFO              = 2,
    GLOBUS_S3_DEBUG_DEBUG             = 4,
    GLOBUS_S3_DEBUG_WARNING           = 8,
    GLOBUS_S3_DEBUG_ERROR             = 16,
    GLOBUS_S3_DEBUG_ALL               = 31
};

extern const char * globus_i_s3_debug_level_names[];


#define GlobusS3Log(level, ...) \
    do { \
        int level__ = level; \
        if (level__ > GLOBUS_S3_DEBUG_ERROR || level__  < 0) { \
            level__ = 1; \
        } \
        GlobusDebugPrintf(GLOBUS_S3, level__, \
            ("s3: %5s: %s: ", \
             globus_i_s3_debug_level_names[level__], __func__)); \
        GlobusDebugPrintf(GLOBUS_S3, level__, \
            (__VA_ARGS__)); \
    } while (0)

#define GlobusS3Trace(...)   GlobusS3Log(GLOBUS_S3_DEBUG_TRACE, __VA_ARGS__)
#define GlobusS3Info(...)    GlobusS3Log(GLOBUS_S3_DEBUG_INFO,  __VA_ARGS__)
#define GlobusS3Debug(...)   GlobusS3Log(GLOBUS_S3_DEBUG_DEBUG, __VA_ARGS__)
#define GlobusS3Warning(...) GlobusS3Log(GLOBUS_S3_DEBUG_WARNING, __VA_ARGS__)
#define GlobusS3Error(...)   GlobusS3Log(GLOBUS_S3_DEBUG_ERROR, __VA_ARGS__) 

#define GlobusS3Enter() GlobusS3Trace("enter\n")
#define GlobusS3ExitInt(i) GlobusS3Trace("exit: %d\n", (i))
#define GlobusS3Exit() GlobusS3Trace("exit\n")
#define GlobusS3ExitStatus(status) GlobusS3Trace("exit status=%d: %s\n", \
                      (status), S3_get_status_name(status))
#define GlobusS3DebugOpStatus(s3_command, state) \
    do { \
        if ((state)->status != S3StatusOK) \
        { \
            GlobusS3Debug(#s3_command ": op=%p status=%s[%d]\n", \
                    (void *) (state)->op, \
                    S3_get_status_name((state)->status), \
                    (state)->status); \
        } \
    } while (0)

#define GlobusS3LogResult(level, result) \
        do { \
            if (result) \
            { \
                char *errstr = globus_error_print_friendly(globus_error_peek(result)); \
                GlobusS3Log(level, "result=%#x message=%s\n", result, errstr); \
                free(errstr); \
            } \
            else \
            { \
                GlobusS3Log(level, "result=GLOBUS_SUCCESS\n"); \
            } \
        } while (0)
#define GlobusS3ExitResult(result) \
        do { \
            if (result) \
            { \
                char *errstr = globus_error_print_friendly(globus_error_peek(result)); \
                GlobusS3Error("result=%#x message=%s\n", result, errstr); \
                free(errstr); \
            } \
            else \
            { \
                GlobusS3Trace("result=GLOBUS_SUCCESS\n"); \
            } \
        } while (0)
#endif /* GLOBUS_GRIDFTP_SERVER_S3_BASE_H */
// vim: ft=c:
