/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file module.c
 * Implementation of the Ceph DSI
 */

#include "globus_i_ceph.h"
#include "globus_gridftp_server_s3_version.h"

#define GLOBUS_L_CEPH "ceph"

char                                   *globus_i_s3_rg_admin_bucket;
char                                   *globus_i_s3_rg_admin_access_key_id;
char                                   *globus_i_s3_rg_admin_secret_access_key;

static
int
globus_l_ceph_activate(void);

static
int
globus_l_ceph_deactivate(void);

static
globus_gfs_storage_iface_t              globus_l_ceph_dsi_iface =
{
    .descriptor                       = GLOBUS_GFS_DSI_DESCRIPTOR_BLOCKING
                              | GLOBUS_GFS_DSI_DESCRIPTOR_SENDER
                              | GLOBUS_GFS_DSI_DESCRIPTOR_REQUIRES_ORDERED_DATA,
    .init_func                        = globus_i_ceph_start,
    .destroy_func                     = globus_i_s3_destroy,
    .send_func                        = globus_i_s3_send,
    .recv_func                        = globus_i_s3_recv,
    .command_func                     = globus_i_s3_command,
    .stat_func                        = globus_i_s3_stat
};

GlobusExtensionDefineModule(globus_gridftp_server_ceph) =
{
    .module_name                      = "globus_gridftp_server_ceph",
    .activation_func                  = globus_l_ceph_activate,
    .deactivation_func                = globus_l_ceph_deactivate,
    .version                          = &local_version
};


static
int
globus_l_ceph_activate(void)
{
    char                               *config_string = NULL;
    int                                 rc = GLOBUS_SUCCESS;
    globus_result_t                     result = GLOBUS_SUCCESS;
    globus_i_s3_config_key_t            ceph_config_keys[] =
    {
        {
            .value_type = GLOBUS_S3_CONFIG_STRING,
            .key_name = "ceph_rgw_admin_bucket",
            .value_string = &globus_i_s3_rg_admin_bucket
        },
        {
            .value_type = GLOBUS_S3_CONFIG_STRING,
            .key_name = "ceph_rgw_admin_access_key_id",
            .value_string = &globus_i_s3_rg_admin_access_key_id,
            .required = true
        },
        {
            .value_type = GLOBUS_S3_CONFIG_STRING,
            .key_name = "ceph_rgw_admin_secret_access_key",
            .value_string = &globus_i_s3_rg_admin_secret_access_key,
            .required = true
        },
        {
            /* This is present in the base config, but marked "required"
             * here for configuration error checking
             */
            .value_type = GLOBUS_S3_CONFIG_STRING,
            .key_name = "host_name",
            .value_string = &globus_i_s3_default_hostname,
            .required = true
        },
        {
            .key_name = NULL
        }
    };

    globus_gridftp_server_get_config_string(NULL, &config_string);

    if (config_string == NULL)
    {
        result = globus_eval_path(
                "${sysconfdir}/globus/globus-gridftp-server-ceph.conf",
                &config_string);
        if (result != GLOBUS_SUCCESS)
        {
            goto done;
        }
    }
    result = globus_i_s3_config_parse(config_string, ceph_config_keys);
    if (result != GLOBUS_SUCCESS)
    {
        goto parse_config_fail;
    }
    result = globus_module_activate(GLOBUS_GRIDFTP_SERVER_S3_BASE_MODULE);

    if (result != GLOBUS_SUCCESS)
    {
        goto module_activate_fail;
    }

    rc = globus_extension_registry_add(
        GLOBUS_GFS_DSI_REGISTRY,
        GLOBUS_L_CEPH,
        GlobusExtensionMyModule(globus_gridftp_server_ceph),
        &globus_l_ceph_dsi_iface);

    if (rc != GLOBUS_SUCCESS)
    {
        globus_module_deactivate(GLOBUS_GRIDFTP_SERVER_S3_BASE_MODULE);
module_activate_fail:
parse_config_fail:
done:
        free(config_string);
    }
    GlobusS3ExitResult(result);
    return result;
}
/* globus_l_ceph_activate() */

static
int
globus_l_ceph_deactivate(void)
{
    globus_extension_registry_remove(GLOBUS_GFS_DSI_REGISTRY, GLOBUS_L_CEPH);

    return globus_module_deactivate(GLOBUS_GRIDFTP_SERVER_S3_BASE_MODULE);
}
/* globus_l_ceph_deactivate() */
// vim: ft=c:
