/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file rename.c
 * Implementation of the s3 DSI rename
 */
#include "globus_gridftp_server_s3_base.h"

static
void *
globus_l_s3_rename_thread(
    void                               *user_arg);

static
void
globus_l_s3_rename_complete(
    S3Status                            status,
    const S3ErrorDetails               *errorDetails,
    void                               *callbackData);

static
S3Status
globus_l_s3_rename_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

typedef struct
{
    /** GridFTP rename operation */
    globus_gfs_operation_t              op;
    /** Bucket Context derived from the session info and the source pathname */
    S3BucketContext                     bucket_context;
    /** From_pathname for the original cmd_info */
    const char                         *orig_from;
    /** Copy of the pathname string split into bucket_name and key */
    char                               *from_pathname;
    /** pointer into pathname that contains the bucket name */
    char                               *from_bucket_name;
    /** pointer into pathname that contains the key name */
    char                               *from_key;
    /** Number of times the current operation has been retried */
    int                                 retries;
    /** S3 Status of the operation */
    S3Status                            status;
    /** Result of the operation */
    globus_result_t                     result;
    /** pointer into to_pathname that contains the bucket name */
    char                               *to_bucket_name;
    /** pointer into to_pathname that contains the key name */
    char                               *to_key;
    char                               *to_pathname;
}
globus_l_s3_rename_t;

/**
 * @brief rename
 * @details
 */
void
globus_i_s3_rename(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_l_s3_rename_t               *rename_state = NULL;
    globus_thread_t                     t;
    int                                 rc;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    GlobusS3Info("rename: op=%p from_pathname=%s pathname=%s\n",
            (void *) op,
            cmd_info->from_pathname,
            cmd_info->pathname);

    if (strcmp(cmd_info->from_pathname, cmd_info->pathname) == 0)
    {
        result = GlobusGFSErrorGeneric("Cannot rename to identical name\n");
        goto failure;
    }
    assert(s3_handle != NULL);

    rename_state = malloc(sizeof(globus_l_s3_rename_t));
    if (rename_state == NULL)
    {
        result = GlobusGFSErrorMemory("rename_state");

        goto failure;
    }

    *rename_state = (globus_l_s3_rename_t)
    {
        .op                           = op,
        .orig_from                    = cmd_info->from_pathname,
        .from_pathname                = strdup(cmd_info->from_pathname),
        .to_pathname                  = strdup(cmd_info->pathname)
    };
    if (rename_state->from_pathname == NULL)
    {
        result = GlobusGFSErrorMemory("from_pathname");

        goto failure;
    }
    if (rename_state->to_pathname == NULL)
    {
        result = GlobusGFSErrorMemory("to_pathname");

        goto failure;
    }

    globus_i_s3_path_parse(rename_state->from_pathname, &rename_state->from_bucket_name, &rename_state->from_key);
    globus_i_s3_path_parse(rename_state->to_pathname, &rename_state->to_bucket_name, &rename_state->to_key);

    if (rename_state->from_key == NULL)
    {
        result = GlobusGFSErrorGeneric("Invalid source filename");

        goto failure;
    }
    else if (rename_state->to_key == NULL)
    {
        result = GlobusGFSErrorGeneric("Invalid destination filename");

        goto failure;
    }

    rename_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, rename_state->from_bucket_name);

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_rename_thread,
            rename_state);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");
failure:
        
        GlobusS3Info("finished_command: op=%p result=%#x\n",
                    (void *) op,
                    (unsigned int) result);
        globus_gridftp_server_finished_command( 
                op,
                result,
                NULL);
        if (rename_state)
        {
            free(rename_state->from_pathname);
            free(rename_state->to_pathname);
            free(rename_state);
        }
    }

    GlobusS3Exit();

    return;
}

static
void *
globus_l_s3_rename_thread(
    void                               *user_arg)
{
    globus_l_s3_rename_t               *rename_state = user_arg;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    assert(rename_state != NULL);

    rename_state->retries = 0;
    do
    {
        rename_state->status = S3StatusOK;
        rename_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(rename_state->retries);

        GlobusS3Info("REST: op=%p S3_copy_object(from_bucket=%s, from_key=%s, to_bucket=%s, to_key=%s) try=%d\n",
                    (void *) rename_state->op,
                    rename_state->from_bucket_name,
                    rename_state->from_key,
                    rename_state->to_bucket_name,
                    rename_state->to_key,
                    rename_state->retries);

        S3_copy_object(&rename_state->bucket_context,
                rename_state->from_key,
                rename_state->to_bucket_name,
                rename_state->to_key,
                NULL,
                NULL,
                0,
                NULL,
                NULL,
                &(S3ResponseHandler)
                {
                    .propertiesCallback   = globus_l_s3_rename_properties,
                    .completeCallback     = globus_l_s3_rename_complete,
                },
                rename_state);

        GlobusS3DebugOpStatus(S3_copy_object, rename_state);
    }
    while (S3_status_is_retryable(rename_state->status)
        && (rename_state->retries++) < globus_i_s3_max_retries);



    if (rename_state->result == GLOBUS_SUCCESS)
    {
        rename_state->retries = 0;
        do
        {
            rename_state->status = S3StatusOK;
            rename_state->result = GLOBUS_SUCCESS;

            globus_i_s3_retry_delay(rename_state->retries);

            GlobusS3Info("REST: op=%p S3_delete_object(bucket=%s, key=%s) try=%d\n",
                    (void *) rename_state->op,
                    rename_state->from_bucket_name,
                    rename_state->from_key,
                    rename_state->retries);

            S3_delete_object(
                    &rename_state->bucket_context,
                    rename_state->from_key,
                    NULL,
                    &(S3ResponseHandler)
                    {
                        .propertiesCallback
                                      = globus_l_s3_rename_properties,
                        .completeCallback
                                      = globus_l_s3_rename_complete,
                    },
                    rename_state);
            GlobusS3DebugOpStatus(S3_delete_object, rename_state);
        }
        while (S3_status_is_retryable(rename_state->status)
            && (rename_state->retries++) < globus_i_s3_max_retries);
    }

    result = rename_state->result;

    GlobusS3Info("finished_command: op=%p result=%#x\n",
                 (void *) rename_state->op,
                 (unsigned int) result);
    globus_gridftp_server_finished_command(rename_state->op, result, NULL);

    free(rename_state->from_pathname);
    free(rename_state->to_pathname);
    free(rename_state);

    return NULL;
}
/* globus_l_s3_rename_thread() */

static
S3Status
globus_l_s3_rename_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_rename_properties() */

static
void
globus_l_s3_rename_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_l_s3_rename_t               *rename_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK)
    {
        rename_state->status = status;
        rename_state->result = GlobusGFSErrorS3(
                    status,
                    rename_state->orig_from,
                    error_details);
    }

    GlobusS3Exit();
}
/* globus_l_s3_rename_complete() */
// vim: ft=c:
